% 
%   MLCV Winter 2015
%   Homework Assignment 03 - Gaussian Mixture Models
%
%   Exercise 2c) EM algorithm
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function [ gamma, M, Sigma, pie, llh ] = mgmmEM( X, K, varargin )
%MGMM Summary of this function goes here
% X - Nxdim matrix, the data: N samples with dim properties each
% K - number of gaussians to assign the data to
%
% gamma - NxK matrix, the responsibilities
% M     - Kxdim matrix, the means
% Sigma - KxKxdim array, the covariances
% pie   - 1xK, the mixing coefficients
% llh   - the final log-likelihood
init = [];
if nargin > 2
    init = varargin{1};
end

plot = 0;
if nargin > 3
    plot = varargin{2};
end

[N, dim] = size(X);
Xmin = min(X);
Xmax = max(X);
Xcov = cov(X);

% Initialize means randomly within the data range
[M, Sigma, pie] = initEM(init);

if plot
    mscatter(X, gmmIdx);
    hold on
    mscatter(M, (1:K)', 'x'); 
end

% Compute initial log-likelihood
llh = computeLogLikelihood(M, Sigma, pie);

% Convergence criteria
eTol = 1e-2; 
converged = 0;
maxIter = 100;

iter = 1;
while ~converged && iter < maxIter
    
    % Store log-likelihood
    llhOld = llh;
    
    % E-step: Compute responsibilities based on current M, Sigma and pie
    gamma = computeResponsibilities(M, Sigma, pie);

    % M-step: Update M, Sigma and pie based on current responsibilities
    [M, Sigma, pie] = updateParams(gamma);
     
    % Evaluate converge criterion
    llh = computeLogLikelihood(M, Sigma, pie);
    if abs(llh - llhOld) < eTol
        converged = true;
    end
    iter = iter + 1;
end


if converged
    fprintf('Converged after %d iterations\n', iter);
else
    fprintf('Did not converge after %d iterations\n', iter);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [M, Sigma, pie] = initEM(initModel)
    
    % Initialize mixing coefficients with equal probability
    pie = ones(1,K) / K;
 
    % Initialize means
    if isempty(initModel.M)
        M = computeRandomMeans();
    else
        M = initModel.M;
    end
    
    % Initialize covariances
    Sigma = zeros(dim, dim, K);
    if isempty(initModel.Sigma)      
        for k=1:K
            Sigma(:,:,k) = Xcov + eye(dim)*eps;    
        end
    else
        Sigma = initModel.Sigma;    
    end    

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Expectation: Compute responsibilities based on current parameters
function gamma = computeResponsibilities(M, Sigma, pie)
    % Compute gaussian pdfs
    NG = zeros(N, K);
    for k=1:K
        NG(:,k) = mvnpdf(X, M(k,:), Sigma(:,:,k)); % N x 1
    end 
    
    % Compute normalizer
    normalizer = sum( bsxfun(@times, pie, NG), 2 ); % N x 1
    
    % Compute responsibilities
    gamma = zeros(N, K);    
    for k=1:K
        gamma(:, k) = pie(k)*NG(:,k) ./ normalizer; % N x 1       
    end 
    
    % Visualize the assignment
    if plot, 
        mscatter(X, gamma);
        legend(['Iter: ' num2str(iter)]);
    end       
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Maximization: Update parameters M, Sigma, pie using current responsibilities
function [M, Sigma, pie] = updateParams(gamma)  
    M = zeros(K, dim);
    Sigma = zeros(dim, dim, K);
    pie = zeros(1, K);
    for k=1:K    
        % Compute normalizer for model k
        Nk = sum(gamma(:,k)); % 1 x 1
        
        % Update mixing coefficient of model k
        pie(k) = Nk / N;
        
        % Update mean of model k
        M(k,:) =  gamma(:,k)'*X / Nk; % 1 x dim
        
        % Update covariance of model k
        dXMk = bsxfun(@minus, X, M(k,:)); % N x dim  
        Sigma(:,:,k) = dXMk' * diag(gamma(:,k)) * dXMk / Nk;
        Sigma(:,:,k) = fixCovariance(Sigma(:,:,k)); %nearestSPD(Sigma(:,:,k));
    end
    
    % Visualize the new means
    if plot, 
        mscatter(M, (1:K)', 'x'); 
    end     
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute the log-likelihood L(theta|x) = P(x|theta)
function llh = computeLogLikelihood(M, Sigma, pie)
    % Compute gaussian pdfs
    NG = zeros(N, K);
    for k=1:K    
        NG(:,k) = mvnpdf(X, M(k,:), Sigma(:,:,k)); % N x 1
    end 
    
    % Compute normalizer
    normalizer = sum( bsxfun(@times, pie, NG), 2 ); % N x 1
    
    llh = sum( log( normalizer ) );
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Compute random means
function M = computeRandomMeans()
    %M = X(randi(N, k, 1), :);
    M = zeros(K,dim);
    for k=1:K
        M(k,:) = Xmin + (Xmax-Xmin).*rand(1,dim);
    end  
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Make sure covariance matrix is symmetric and positive definite
function C = fixCovariance(C)

    % Avoid zeros/NaN errors
    if isequal(C, zeros(dim)) || any(any(isnan(C)))
        C = Xcov;
    end
    
    % Avoid non-positive definite covariance matrices
    [V, D] = eig(C); 
    if D(1,1) <= 0
        C = C + V(:,1)*V(:,1)'*(eps(D(1,1))-D(1,1)); 
    end    
    
    % Symmetrize
    if ~issymmetric(C)
        C = (C + C') / 2;
    end 
    
    C = C + eps*eye(dim);
end

end
